import numpy as np
import scipy.linalg as la
# dynamical lib
from st_dsne import ST_DSNE


def DSNE(
    data,
    V,
    Y,
    pca_d=None,
    perplexity=3.0,
    K = 16,
    threshold_V=1e-9,
    separate_threshold=1e-9,
    max_iter=1000,
    mom_switch_iter=250,
    momentum=0.5,
    final_momentum=0.8,
    eta=0.1,
    epsilon_kl=1e-9,
    epsilon_dsne=1e-9,
    with_norm=True,
    seed=None,
    random_state=None,
    copy_data=True,
    verbose=False,
):
    """
    Run DSNE on data (data points), V (velocity), Y (map points).

    @param data         The data points matrix, with shape (N,D).

    @param V            The velocity matrix, with shape (N,D), each V_i with shape D is the velocity of point data_i.

    @param Y            The low dimension map points matrix, with shape (N,d), each Y_i with shape d is the map points of point data_i.

    @param pca_d        The dimensionality of data and V is reduced via PCA
                        to this dimensionality, when D>5000, can use pca to reduce the dimension of data and V.

    @param perplexity   The perplexity controls the effective number of
                        neighbors.

    @param K            The number of KNN neighbors, which is used to compute the similarity matrix P and Q.

    @param threshold_V  A threshold. When max(abs(V_i)) < thershold_V, will omit to find its velocity embedding and set w_i=0

    @param separate_threshold.  A threshold. When we find the KNN neighbor of data_i, only when j satisfies
                            || data_j - data_i||^2 > separate_threshold and max( abs(Y_j- Y_i) ) > separate_threshold
                        will be considered as the neighbor of data_i.

    @param max_iter     The max number of iteration to update the velocity embedding W

    @mom_switch_iter    The momentum swith times. On iter in  [0, mom_switch_iter) will use momentum, on iter
                        in [mom_switch_iter, max_iter) will use final_momentum.

    @momentum           The initial momentum scalar.

    @final_momentum     The final momentum scalar.

    @eta                Learning rate. Scalar.

    @epsilon_kl         The threshold to check the convergence of velocity embedding for fixed \beta_Q each round

    @epsilon_dsne       The threshold to check the convergence of DSNE.

    @with_norm          Bool. If true, return the velocity embedding W wiht approximate norm based on data,V and Y;
                        If false, return velocity embddding W with unit length for each W_i, i.e ||W_i|| =1

    @seed               Int. The random seed.

    @param random_state A numpy RandomState object; if None, use
                        the numpy.random singleton. Init the RandomState
                        with a fixed seed to obtain consistent results
                        from run to run.

    @param copy_data    Copy the data to prevent it from being modified
                        by the C code.

    @param verbose      Verbose output from the training process.
    """
    N, D = data.shape
    N_V, D_V = V.shape
    N_Y, d = Y.shape
    if(not (N== N_Y and  N_Y == N_V and D ==D_V)):
        print("The shape of data: {}, V: {}, Y: {} mismatch!".format(data.shape, V.shape, Y.shape))
        raise


    if pca_d is None:
        if copy_data:
            X = np.copy(data)
            V = np.copy(V)
            Y = np.copy(Y)
        else:
            X = data
    else:
        # do PCA
        # data -= data.mean(axis=0)

        # working with covariance + (svd on cov.) is
        # much faster than svd on data directly.
        cov = np.dot(data.T, data) / N
        u, s, v = la.svd(cov, full_matrices=False)
        u = u[:, 0:pca_d]
        X = np.dot(data, u)
        V = np.dot(V, u)
    if seed is None:
        if random_state is None:
            seed = np.random.randint(2 ** 32 - 1)
        else:
            seed = random_state.randint(2 ** 32 - 1)

    if K is None:
        K = int(3*perplexity)

    if K > N - 1:
        if verbose:
            print("K: {} is too large, set it to N - 1".format(K, N-1))
        K = N - 1

    st_dsne= ST_DSNE()
    original_dtype = X.dtype
    dtype_changed = False
    if X.dtype is not np.dtype(np.float64):
        dtype_changed = True
        X = np.asarray(X, dtype=np.float64)
    if V.dtype is not np.dtype(np.float64):
        V = np.asarray(V, dtype=np.float64)
        dtype_changed = True
    if Y.dtype is not np.dtype(np.float64):
        Y = np.asarray(Y, dtype=np.float64)
        dtype_changed = True
    W = st_dsne.run(X, V, Y, N, K, X.shape[1], d, perplexity, threshold_V,
                    separate_threshold,  max_iter, mom_switch_iter, momentum, final_momentum, eta,epsilon_kl,
                        epsilon_dsne, with_norm,
    seed, verbose)
    if dtype_changed:
        W = np.asarray(W, dtype=original_dtype)
    return W


def DSNE_approximate(
    data,
    V,
    Y,
    pca_d=None,
    perplexity=3.0,
    K = 16,
    threshold_V=1e-9,
    separate_threshold=1e-9,
    with_norm = True,
    seed=None,
    random_state=None,
    copy_data=True,
    verbose=False,
):
    """
        Run DSNE approximately on data (data points), V (velocity), Y (map points).

        @param data         The data points matrix, with shape (N,D).

        @param V            The velocity matrix, with shape (N,D), each V_i with shape D is the velocity of point data_i.

        @param Y            The low dimension map points matrix, with shape (N,d), each Y_i with shape d is the map points of point data_i.

        @param pca_d        The dimensionality of data and V is reduced via PCA
                            to this dimensionality, when D>5000, can use pca to reduce the dimension of data and V.

        @param perplexity   The perplexity controls the effective number of
                            neighbors.

        @param K            The number of KNN neighbors, which is used to compute the similarity matrix P and Q.

        @param threshold_V  A threshold. When max(abs(V_i)) < thershold_V, will omit to find its velocity embedding and set w_i=0

        @param separate_threshold.  A threshold. When we find the KNN neighbor of data_i, only when j satisfies
                                || data_j - data_i||^2 > separate_threshold and max( abs(Y_j- Y_i) ) > separate_threshold
                            will be considered as the neighbor of data_i.

        @with_norm          Bool. If true, return the velocity embedding W wiht approximate norm based on data,V and Y;
                            If false, return velocity embddding W with unit length for each W_i, i.e ||W_i|| =1

        @seed               Int. The random seed.

        @param random_state A numpy RandomState object; if None, use
                            the numpy.random singleton. Init the RandomState
                            with a fixed seed to obtain consistent results
                            from run to run.

        @param copy_data    Copy the data to prevent it from being modified
                            by the C code.

        @param verbose      Verbose output from the training process.
        """

    N, D = data.shape
    N_V, D_V = V.shape
    N_Y, d = Y.shape
    if (not (N == N_Y and N_Y == N_V and D == D_V)):
        print("The shape of data: {}, V: {}, Y: {} mismatch!".format(data.shape, V.shape, Y.shape))
        raise


    if pca_d is None:
        if copy_data:
            X = np.copy(data)
            V = np.copy(V)
            Y = np.copy(Y)
        else:
            X = data
    else:
        # do PCA
        data -= data.mean(axis=0)

        # working with covariance + (svd on cov.) is
        # much faster than svd on data directly.
        cov = np.dot(data.T, data) / N
        u, s, v = la.svd(cov, full_matrices=False)
        u = u[:, 0:pca_d]
        X = np.dot(data, u)
        V = np.dot(V, u)
    if seed is None:
        if random_state is None:
            seed = np.random.randint(2 ** 32 - 1)
        else:
            seed = random_state.randint(2 ** 32 - 1)
    if K is None:
        K = int(6*perplexity)

    if K > N - 1:
        if verbose:
            print("K: {} is too large, set it to N - 1".format(K, N-1))
        K = N - 1
    st_dsne = ST_DSNE()
    original_dtype = X.dtype
    dtype_changed = False
    if X.dtype is not np.dtype(np.float64):
        dtype_changed = True
        X = np.asarray(X,dtype=np.float64)
    if V.dtype is not np.dtype(np.float64):
        V = np.asarray(V,dtype=np.float64)
        dtype_changed = True
    if Y.dtype is not np.dtype(np.float64):
        Y = np.asarray(Y,dtype=np.float64)
        dtype_changed = True
        # X, V, N, D, d, alpha, perplexity, theta, seed, verbose = False):
    # X, V, Y, N_full, D, d, perplexity, max_iter, mom_switch_iter,
    # momentum, final_momentum, eta, exact, seed, verbose
    # run(self, X, V, Y, N_full, D, d, perplexity, max_iter, threshold_V,
    #     separate_threshold, mom_switch_iter,
    #     momentum, final_momentum, eta, exact, seed, verbose=False)
    # print("perplexity", perplexity)
    W = st_dsne.run_approximate(X, V, Y, N, K, X.shape[1], d, perplexity, threshold_V,
                    separate_threshold, with_norm,
            seed, verbose)
    if dtype_changed:
        W = np.asarray(W, dtype=original_dtype)
    return W



import numpy as np
import scvelo as scv
from scipy.sparse import issparse
from dsne import DSNE, DSNE_approximate


scv.settings.verbosity = 3  # show errors(0), warnings(1), info(2), hints(3)
scv.settings.presenter_view = True  # set max width size for presenter view
scv.settings.set_figure_params('scvelo')  # for beautified visualization
adata = scv.datasets.pancreas()

scv.pp.filter_and_normalize(adata, min_shared_counts=20, n_top_genes=2000)
scv.pp.moments(adata, n_pcs=30, n_neighbors=30)

scv.tl.velocity(adata)


def get_X_V_Y(adata,vkey="velocity",
              xkey="Ms",
              basis=None,
              gene_subset=None,
              ):

    subset = np.ones(adata.n_vars, bool)
    if gene_subset is not None:
        var_names_subset = adata.var_names.isin(gene_subset)
        subset &= var_names_subset if len(var_names_subset) > 0 else gene_subset
    elif f"{vkey}_genes" in adata.var.keys():
        subset &= np.array(adata.var[f"{vkey}_genes"].values, dtype=bool)

    xkey = xkey if xkey in adata.layers.keys() else "spliced"
    basis = 'umap' if basis is None else basis
    X = np.array(
        adata.layers[xkey].A[:, subset]
        if issparse(adata.layers[xkey])
        else adata.layers[xkey][:, subset]
    )
    V = np.array(
        adata.layers[vkey].A[:, subset]
        if issparse(adata.layers[vkey])
        else adata.layers[vkey][:, subset]
    )
    # V -= np.nanmean(V, axis=1)[:, None]
    Y =np.array(
        adata.obsm[f"X_{basis}"]
    )


    nans = np.isnan(np.sum(V, axis=0))
    if np.any(nans):
        X = X[:, ~nans]
        V = V[:, ~nans]
    return X,V,Y


X,V,X_2d = get_X_V_Y(adata,vkey="velocity",xkey="Ms",basis="umap")


V_2d = DSNE(X, V, Y=X_2d,
            perplexity=3.0,
            K=16,
            threshold_V=1e-8,
            separate_threshold=1e-8,
            max_iter=600,
            mom_switch_iter=250,
            momentum=0.5,
            final_momentum=0.8,
            eta=0.1,
            epsilon_kl=1e-16,
            epsilon_dsne=1e-16,
            seed=6,
            random_state=None,
            copy_data=False,
            with_norm=True,
            verbose=True)

adata.obsm["X_DSNE"] = X_2d
adata.obsm["V_DSNE"] = V_2d
title ="DSNE"
scv.pl.velocity_embedding_stream(adata, title=title, basis='umap',V=adata.obsm["V_DSNE"], smooth=0.5,density=2,)

scv.pl.velocity_embedding_grid(adata, title=title, basis='umap',V=adata.obsm["V_DSNE"], smooth=0.5,density=2,)


scv.pl.velocity_embedding(adata,  title=title,basis='umap',V = adata.obsm["V_DSNE"])


title ="DSNE_approximate"
V_2d = DSNE_approximate(X, V, Y=X_2d,
                        perplexity=3.0,
                        K=16,
                        threshold_V=1e-8,
                        separate_threshold=1e-8,
                        seed=6,
                        random_state=None,
                        copy_data=False,
                        with_norm=True,
                        verbose=True)

adata.obsm["X_DSNE_approximate"] = X_2d
adata.obsm["V_DSNE_approximate"] = V_2d

scv.pl.velocity_embedding_stream(adata, basis='umap',V=adata.obsm["V_DSNE_approximate"],  title=title, smooth=0.5,density=2,)

scv.pl.velocity_embedding_grid(adata, basis='umap',V=adata.obsm["V_DSNE_approximate"],  title=title, smooth=0.5,density=2,)

scv.pl.velocity_embedding(adata, basis='umap',V = adata.obsm["V_DSNE_approximate"], title=title)



import os
import numpy as np
import umap
from anndata import AnnData, read_h5ad
from matplotlib import pyplot as plt
from tsne import bh_sne
from dsne import DSNE, DSNE_approximate, scvelo_velocity_embedding


def unitLength(V):
    V_ = V/np.sqrt(np.sum(V*V,axis=1,keepdims=True))
    return V_
def velocity_accuracy(V, V_exact, N):
    index = np.asarray(list(range(N-1))+ list(np.arange(N, 2*N-1)) + list(np.arange(2*N, 3*N-1)), dtype=np.int)
    V_ = V[index,:]
    V_exact_ = V_exact[index,:]
    V_unit = unitLength(V_)
    V_exact_unit = unitLength(V_exact_)
    accu = np.sum( V_unit* V_exact_unit )/(V_.shape[0]*1.)
    return accu

def simulate_data_base(N = 100, D = 30, theta=0.5, perplexity=20,  nrep=1, resimulate =False, save =True, file_prefix = "./data/"):
    # 3 types
    file_base = f"unexact_simulation_data_N_{N}_D_{D}_nrep_{nrep}.h5ad"
    file = os.path.join(file_prefix, file_base)
    if os.path.exists(file) and not resimulate:
        adata = read_h5ad(file)
        return adata

    V = np.abs(np.random.randn(*(N * 3, D)))*6
    x_1 = np.asarray([0,]*D)
    x_2 =  np.asarray([50,]*D)
    x_3 = np.asarray([160, ]*D)
    X = np.zeros_like(V)
    X[0, :] = x_1
    X[N, :] = x_2
    X[N*2, :] = x_3
    for i in np.arange(N-1):
        X[i+1,:] = X[i,:] + V[i,:]
        X[i + N +1, :] = X[i+ N, :] + V[i+ N, :]
        X[i + N*2 + 1, :] = X[i + N*2, :] + V[i + N*2, :]

    y = np.asarray([0,] *N + [1,]  * N + [ 2,]*N)
    X_tsne = bh_sne(X,copy_data=True,perplexity=perplexity, theta=theta, verbose=True)
    X_umap = umap.UMAP(n_components=2).fit_transform(X)
    X_tsne = np.asarray(X_tsne,dtype=np.float64)
    X_umap = np.asarray(X_umap, dtype=np.float64)
    V_2d_tsne_true = np.zeros_like(X_tsne)
    V_2d_umap_true = np.zeros_like(X_umap)
    for i in range(V_2d_tsne_true.shape[0]-1):
        V_2d_tsne_true[i,:] = X_tsne[i+1,:] - X_tsne[i,:]
        V_2d_umap_true[i,:] = X_umap[i+1,:] - X_umap[i,:]
    adata = AnnData(X=X,layers = {"V":V}, obsm={"X_umap":X_umap,"X_tsne":X_tsne,
                                                "V_2d_tsne_true": V_2d_tsne_true,
                                                "V_2d_umap_true": V_2d_umap_true,
                                                }, obs={"clusters":y})
    if save:
        adata.write_h5ad(file)
    return adata


nrep = 6
N = 50
D = 30
K = 6
perplexity = 1
theta = 0.5
exact = False
basis = "tsne"
# basis ="umap"

adata = simulate_data_base(N = N, D = D, nrep=nrep,theta=theta, resimulate =False, save =False, file_prefix = "./data/")
X = adata.X
V = adata.layers["V"]
X_2d = adata.obsm[f"X_{basis}"]
y =  np.asarray([0,] *N + [1,]  * N + [ 2,]*N)
adata.obs['clusters'] = y
V_2d_true = adata.obsm[f"V_2d_{basis}_true"]

V_2d = DSNE(X, V, Y=X_2d,
            K=K,
            perplexity=perplexity,
            verbose=True)

plt.scatter(X_2d[:, 0], X_2d[:, 1], c=y,s=0.1)
plt.quiver(X_2d[:, 0], X_2d[:, 1],V_2d[:,0], V_2d[:,1],y,
           # width = 1,
           lw=0.01,
           headwidth=3,
           headlength=2,
           headaxislength =1,
           # scale=1
           )
# plt.legend("0","1","2")
for i in np.arange(N):
    plt.text(X_2d[i, 0], X_2d[i, 1], s=eval(f"i"))
    plt.text(X_2d[i+N, 0], X_2d[i+N, 1], s=eval(f"i"))
    plt.text(X_2d[i + N*2, 0], X_2d[i + N*2, 1], s=eval(f"i"))

plt.colorbar()
accu = velocity_accuracy(V_2d, V_2d_true, N)
basis_name = "UMAP" if basis == 'umap' else "t-SNE"
title = "DSNE velocity embedding on {} with accuracy {:5.3f}".format(basis_name, accu)
print(title)
plt.title(title)
plt.show()


V_2d = DSNE_approximate(X, V, Y=X_2d,
                        perplexity=perplexity,
                        K=K,
                        verbose=True)

plt.scatter(X_2d[:, 0], X_2d[:, 1], c=y,s=0.1)
plt.quiver(X_2d[:, 0], X_2d[:, 1],V_2d[:,0], V_2d[:,1],y,
           # width = 1,
           lw=0.01,
           headwidth=3,
           headlength=2,
           headaxislength =1,
           # scale=1
           )
plt.colorbar()

for i in np.arange(N):
    plt.text(X_2d[i, 0], X_2d[i, 1], s=eval(f"i"))
    plt.text(X_2d[i+N, 0], X_2d[i+N, 1], s=eval(f"i"))
    plt.text(X_2d[i + N*2, 0], X_2d[i + N*2, 1], s=eval(f"i"))

accu = velocity_accuracy(V_2d, V_2d_true, N)
print(f"DSNE_approximate accu {accu} \n")

plt.title(f"DSNE_approximate, accu {accu}")

plt.show()

V_2d = scvelo_velocity_embedding(X, V, X_2d)
plt.scatter(X_2d[:, 0], X_2d[:, 1], c=y,s=0.1)
plt.quiver(X_2d[:, 0], X_2d[:, 1],V_2d[:,0], V_2d[:,1],y,
           lw=0.01,
           headwidth=3,
           headlength=2,
           headaxislength =1,
           )
plt.colorbar()

# plt.legend("0","1","2")
for i in np.arange(N):
    plt.text(X_2d[i, 0], X_2d[i, 1], s=eval(f"i"))
    plt.text(X_2d[i+N, 0], X_2d[i+N, 1], s=eval(f"i"))
    plt.text(X_2d[i + N*2, 0], X_2d[i + N*2, 1], s=eval(f"i"))

accu = velocity_accuracy(V_2d, V_2d_true, N)
basis_name = "UMAP" if basis == 'umap' else "t-SNE"
title = "scVelo velocity embedding on {} with accuracy {:5.3f}".format(basis_name, accu)
print(title)
plt.title(title)

plt.show()



try:
    from ._generated_version import version as __version__
except ImportError:
    # Package is not installed, parse git tag at runtime
    try:
        import setuptools_scm

        # Code duplicated from setup.py to avoid a dependency on each other
        def parse_git(root, **kwargs):
            """
            Parse function for setuptools_scm
            """
            from setuptools_scm.git import parse

            kwargs[
                "describe_command"
            ] = "git describe --dirty --tags --long --match '*[0-9]*'"
            return parse(root, **kwargs)

        __version__ = setuptools_scm.get_version("./", parse=parse_git)
    except ImportError:
        __version__ = None


from dsne._dsne import DSNE  # noqa
from dsne._dsne import DSNE_approximate  # noqa
from dsne.utils import scvelo_velocity_embedding



def test_import():
    import dsne

    assert dsne.__version__ is not None
    assert dsne.__version__ != "0.0.0"
    assert len(dsne.__version__) > 0


def test_import_st_dsne():
    from dsne import dsne  # noqa
    from dsne import dsne_approximate  # noqa



# cython: linetrace=True
# distutils: define_macros=CYTHON_TRACE=1
# distutils: language=c++

import numpy as np
cimport numpy as np
cimport cython

cdef extern from "dsne.h":
    cdef cppclass DSNE:
        DSNE()
        void run(double* X, double* V, int N_full, int K, int D, double* Y, double* W_full, int no_dims,
                double perplexity, double threshold_V, double separate_threshold,  int max_iter,
                int mom_switch_iter, double momentum, double final_momentum,  double eta, double epsilon_kl, double epsilon_dsne,
                 int with_norm,
                unsigned int seed,  int verbose);
        void run_approximate(double* X, double* V, int N_full, int K, int D, double* Y, double* W_full, int no_dims,
            double perplexity, double threshold_V, double separate_threshold,
            int with_norm,
            unsigned int seed, int verbose);

cdef class ST_DSNE:
    cdef DSNE* thisptr # hold a C++ instance

    def __cinit__(self):
        self.thisptr = new DSNE()

    def __dealloc__(self):
        del self.thisptr

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def run(self, X, V, Y, N_full, K, D, d, perplexity, threshold_V,
                    separate_threshold, max_iter,   mom_switch_iter,
        momentum, final_momentum, eta, epsilon_kl, epsilon_dsne, with_norm,  seed, verbose=False):
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] _X = np.ascontiguousarray(X)
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] _V = np.ascontiguousarray(V)
        # cdef np.ndarray[np.float64_t, ndim=2, mode='c'] Y = np.zeros((N_full, d), dtype=np.float64)
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] _Y = np.ascontiguousarray(Y)
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] W_full= np.zeros((N_full, d), dtype=np.float64)
        self.thisptr.run(&_X[0,0], &_V[0,0], N_full, K, D, &_Y[0,0], &W_full[0,0], d, perplexity, threshold_V,
                    separate_threshold,  max_iter,
                mom_switch_iter, momentum, final_momentum,  eta, epsilon_kl, epsilon_dsne,
                with_norm,  seed,   verbose)
        return W_full

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def run_approximate(self, X, V, Y, N_full, K, D, d, perplexity, threshold_V,
                    separate_threshold, with_norm,  seed, verbose=False):
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] _X = np.ascontiguousarray(X)
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] _V = np.ascontiguousarray(V)
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] _Y = np.ascontiguousarray(Y)
        cdef np.ndarray[np.float64_t, ndim=2, mode='c'] W_full= np.zeros((N_full, d), dtype=np.float64)
        self.thisptr.run_approximate(&_X[0,0], &_V[0,0], N_full, K, D, &_Y[0,0], &W_full[0,0], d, perplexity, threshold_V,
                    separate_threshold,
                    with_norm,
                    seed, verbose)
        return W_full



import os
import platform
import sys

import Cython
import pkg_resources
from Cython.Build import cythonize
from Cython.Distutils import build_ext as _build_ext
from setuptools import Extension, find_packages, setup

# todo 'license_file' option was not specified

if Cython.__version__ < "0.29":
    raise Exception("Please upgrade to Cython 0.29 or newer")

setup_dir = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    filepath = os.path.join(setup_dir, filename)
    with open(filepath) as file:
        return file.read()


class build_ext(_build_ext):
    def build_extensions(self):
        print("Running custom build_ext")

        numpy_incl = pkg_resources.resource_filename("numpy", "core/include")

        if sys.platform == "darwin":
            # Platform: Mac OS
            version, _, _ = platform.mac_ver()
            parts = version.split(".")
            major = int(parts[0])
            minor = int(parts[1])
            patch = int(parts[2]) if len(parts) == 3 else None

            if minor >= 15:
                # Greater than Mac OS: 10.15
                extra_compile_args = [
                    "-I/Library/Developer/CommandLineTools/SDKs/MacOSX.sdk/System/Library/Frameworks/Accelerate.framework/Versions/Current/Frameworks/vecLib.framework/Headers/"
                ]
            elif minor >= 10:
                # Greater than Mac OS: 10.10
                extra_compile_args = [
                    "-I/System/Library/Frameworks/Accelerate.framework/Versions/A/Frameworks/vecLib.framework/Versions/Current/Headers"
                ]
            else:
                extra_compile_args = [
                    "-I/System/Library/Frameworks/vecLib.framework/Headers"
                ]

            ext_module = Extension(
                name="st_dsne",
                sources=[
                    "dsne/includes/st_dsne/dsne.cpp",
                    "dsne/st_dsne.pyx",
                ],
                include_dirs=["dsne/includes/st_dsne/"],
                extra_compile_args=extra_compile_args,
                extra_link_args=["-Wl,-framework", "-Wl,Accelerate", "-lcblas"],
                language="c++",
            )

            self.extensions.append(ext_module)
        else:
            # Platform: Linux
            extra_link_args = ["-lcblas"]

            ext_module = Extension(
                name="st_dsne",
                sources=[
                    "dsne/includes/st_dsne/dsne.cpp",
                    "dsne/st_dsne.pyx",
                ],
                include_dirs=["/usr/local/include", "dsne/includes/st_dsne/"],
                library_dirs=["/usr/local/lib"],
                extra_compile_args=["-msse2", "-O3", "-fPIC", "-w"],
                extra_link_args=extra_link_args,
                language="c++",
            )
            self.extensions.append(ext_module)

        self.extensions = [ext for ext in self.extensions if ext.name != "__dummy__"]

        for ext in self.extensions:
            if hasattr(ext, "include_dirs") and numpy_incl not in ext.include_dirs:
                ext.include_dirs.append(numpy_incl)

        _build_ext.build_extensions(self)


setup(
    name="dsne",
    use_scm_version=True,
    packages=find_packages(),
    # package_dir={"": "src"},
    zip_safe=False,
    include_package_data=True,
    package_data={"dsne": ["includes/*"]},
    # data_files=data_files,
    # Dummy extension to trigger build_ext
    ext_modules=[Extension("__dummy__", sources=[])],
    cmdclass={"build_ext": build_ext},
    # entry_points = {},
    options={"bdist_wheel": {"universal": "1"}},
    python_requires=">=3.6",
    setup_requires=["setuptools_scm"],
    install_requires=read_file("requirements-package.txt").splitlines(),
    extras_require={
        "test": ["pytest"],
        "dev": read_file("requirements.txt").splitlines(),
    },
    description="DSNE algorithms",
    long_description=read_file("README.md"),
    long_description_content_type="text/markdown",
    license="Apache License, Version 2.0",
    maintainer="Songting Shi",
    maintainer_email="songtingstone@gmail.com",
    url="https://github.com/songtingstone/dsne",
    keywords=["DSNE", "algorithms", "numpy", "cython"],
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import os
import numpy as np
import scvelo as scv
from anndata import AnnData, read_h5ad
from dsne import DSNE, DSNE_approximate



N=500
D=300
d=2
K=16
perplexity =6
n_rep=1
exact = False
with_norm = True
basis = 'exact_embeddings'
verbose = False

def unitLength(V):
    V_ = V/np.sqrt(np.sum(V*V,axis=1,keepdims=True))
    return V_
def velocity_accuracy(V, V_exact):
    V_unit = unitLength(V)
    V_exact_unit = unitLength(V_exact)
    accu = np.sum( V_unit* V_exact_unit )/(V.shape[0]*1.)
    return accu

def simulate_data(N=50, D=3, d=2, save =True, file_name_prefix ="./data" ):
    if not os.path.exists(file_name_prefix):
        print("Directory: {} do not exist, create it! \n".format(os.path.abspath(file_name_prefix)))
        os.makedirs(os.path.abspath(file_name_prefix))
    V_2d = np.random.randn(*(N * 3, d)) * 6
    err_2d = np.random.randn(*(N * 3, d))*2
    x_1 = np.asarray([0, ] * d)
    x_2 = np.asarray([50, ] * d)
    x_3 = np.asarray([160, ] * d)
    X_2d = np.zeros_like(V_2d)
    X_2d[0, :] = x_1
    X_2d[N, :] = x_2
    X_2d[N * 2, :] = x_3
    for i in np.arange(N - 1):
        X_2d[i + 1, :] = X_2d[i, :] + V_2d[i, :] + err_2d[i,:]
        X_2d[i + N + 1, :] = X_2d[i + N, :] + V_2d[i + N, :] + err_2d[i + N, :]
        X_2d[i + N * 2 + 1, :] = X_2d[i + N * 2, :] + V_2d[i + N * 2, :] +  err_2d[i + N * 2, :]


    y = np.asarray([0, ] * N + [1, ] * N + [2, ] * N)
    U = np.array(np.random.randn(*(d, D)))
    X = X_2d.__matmul__(U)
    V = V_2d.__matmul__(U)


    adata = AnnData(X=X, layers={"velocity": V},obs={"clusters": y}, obsm={"X_exact_embeddings":X_2d, "V_exact_embeddings":V_2d})
    if save:
        file_name = file_name_prefix+"simulated_data_N_{}_D_{}.h5hd".format(N,D)
        adata.write_h5ad(file_name)
    return adata


adata = simulate_data(N=N,D=D,d=d,save=False)
X = adata.X
V = adata.layers["velocity"]
X_basis = f"X_{basis}"

for method in ['DSNE', "DSNE_approximate","scvelo_velocity_original"  ]:
    X = np.asarray(X, dtype=np.float64)
    V = np.asarray(V, dtype=np.float64)
    Y = None
    if (X_basis in adata.obsm.keys()) and adata.obsm[X_basis] is not None:
        Y = adata.obsm[f"X_{basis}"]

    if Y is None:
        print("Do not get the low dimesnional embedding Y! \n")
        # raise

    Y = np.asarray(Y, dtype=np.float64)
    if method == "DSNE":
        W = DSNE(X, V, Y=Y,
                 K= K,
                 perplexity=perplexity,
                 pca_d=None,
                 threshold_V=1e-8,
                 separate_threshold=1e-8,
                 max_iter=1000,
                 mom_switch_iter=250,
                 momentum=0.5,
                 final_momentum=0.8,
                 eta=0.1,
                 epsilon_kl=1e-16,
                 epsilon_dsne=1e-16,
                 with_norm=with_norm,
                 seed=16,
                 random_state=None,
                 copy_data=True,
                 verbose=verbose)
        vkey = "velocity_dsne"
    elif method == "DSNE_approximate":
        W = DSNE_approximate(X, V, Y=Y,
                                      perplexity=perplexity,
                                      pca_d=None,
                                      threshold_V=1e-8,
                                      separate_threshold=1e-8,
                                      seed=16,
                                      random_state=None,
                                      copy_data=False,
                                      verbose=verbose)
        vkey = "velocity_scvelo"
    elif method == "scvelo_velocity_original":
        adata_tmp = AnnData(X=X, obsm={"X_umap": Y}, layers={"velocity": V, "spliced": X})
        scv.tl.velocity_graph(adata_tmp, xkey='spliced')
        scv.tools.velocity_embedding(adata_tmp, basis="umap")
        W = adata_tmp.obsm["velocity_umap"]
        vkey = "velocity_scvelo_original"
    else:
        print("method: {} do not implemented!\n".format(method))
        # raise
    str_exact = "exact" if exact else "approx"
    adata.obsm[f"{vkey}_{str_exact}_{basis}"] = W
    W_exact = adata.obsm["V_exact_embeddings"]
    accu = velocity_accuracy(W, W_exact)
    print(f"  {method}, {str_exact},  accu: {accu}\n")
    if method == 'DSNE':
        method_str = "DSNE"
    elif method == 'scvelo_velocity_original':
        method_str = "scVelo"
    elif method =="DSNE_approximate":
        method_str = "DSNE-approximate"
    title = "{} on exact embeddings with accuracy {:5.3f}".format(method_str, accu)
    scv.pl.velocity_embedding(adata, basis=basis, V=W, title=title,density=2,)
    scv.pl.velocity_embedding_stream(adata, basis=basis, V=W, title=title,density=2,)
    scv.pl.velocity_embedding_grid(adata, basis=basis, V=W, title=title,)



def test_seed():
    from dsne import dsne
    from sklearn.datasets import load_iris
    from sklearn.decomposition import PCA
    import numpy as np

    iris = load_iris()

    X = iris.data
    V = iris.data
    # y = iris.target
    pca = PCA(n_components=2)
    Y = pca.fit_transform(X)


    t1 = dsne(X,V,Y, random_state=np.random.RandomState(0), copy_data=True)
    # t2 = dsne(X,V,Y, random_state=np.random.RandomState(0), copy_data=True)

    assert t1.shape[0] == 150
    assert t1.shape[1] == 2



import scvelo as scv
from anndata import AnnData


def scvelo_velocity_embedding(X, V, Y):
    adata_tmp = AnnData(X=X,obsm={ "X_umap":Y}, layers={"velocity":V, "spliced":X})
    scv.pp.moments(adata_tmp, n_pcs=30, n_neighbors=30)
    scv.tl.velocity_graph(adata_tmp, xkey='spliced')
    scv.tools.velocity_embedding(adata_tmp,basis="umap")
    W = adata_tmp.obsm["velocity_umap"]
    return W
